﻿using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework.Content;

namespace AnimationClassLibrary
{
    /// <summary>    
    /// 动画片段类，这个类保存一个关键帧集合和动画持续时间。
    /// </summary>
    public class AnimationClip
    {
        /// <summary>
        /// 动画片段的播放长度
        /// </summary>
        [ContentSerializer]
        public TimeSpan Duration { get; private set; }
        
        /// <summary>
        /// 关键帧集合
        /// </summary>
        [ContentSerializer(CollectionItemName = "Keyframe")]
        public List<Keyframe> Keyframes { get; private set; }

        /// <summary>
        /// 创建一个动画片段
        /// </summary>
        /// <param name="Keyframes">关键帧集合</param>
        public AnimationClip(List<Keyframe> Keyframes)
        {
            this.Keyframes = Keyframes;
            //对关键帧根据时间先后进行排序
            Keyframes.Sort(CompareKeyframeTimes);
            // 动画播放的时间就是最后一个关键帧的Time属性
            this.Duration = Keyframes[Keyframes.Count - 1].Time;
        }
        
        // 必须要有一个不带参数的构造函数用于串行化
        private AnimationClip(){}

        // 按照关键帧的时间属性升序排序的Comparison方法
        static int CompareKeyframeTimes(Keyframe a, Keyframe b)
        {
            return a.Time.CompareTo(b.Time);
        }
    }
}
