﻿using System;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;

namespace AnimationClassLibrary
{    
    /// <summary>
    /// 关键帧类，表示一个动画片段（AnimationClip）中一个指定时刻的状态（包含位置、旋转、缩放信息），
    /// 多个关键帧组成一个动画片段。
    /// </summary>
    public class Keyframe
    {
        /// <summary>
        /// 关键帧对应的bone索引
        /// </summary>
        [ContentSerializer]
        public int BoneIndex { get; private set; }

        /// <summary>
        /// 此关键帧离开动画开始时刻的时间
        /// </summary>
        [ContentSerializer]
        public TimeSpan Time { get; private set; }
        
        /// <summary>
        /// 位置
        /// </summary>
        [ContentSerializer]
        public Vector3 Position{ get;internal set;}
        
        /// <summary>
        /// 旋转，使用四元数表示
        /// </summary>
        [ContentSerializer]
        public Quaternion Rotation { get; internal set; }
        
        /// <summary>
        /// 缩放
        /// </summary>
        [ContentSerializer]
        public Vector3 Scale { get; internal set; } 

        /// <summary>
        /// 创建一个关键帧
        /// </summary>
        public Keyframe(int boneIndex,TimeSpan time, Vector3 position, Quaternion rotation, Vector3 scale)
        {
            this.BoneIndex = boneIndex;
            this.Time = time;
            this.Position = position;
            this.Rotation = rotation;
            this.Scale = scale;            
        }

        // 必须要有一个不带参数的构造函数用于串行化
        private Keyframe() { }
    }
}
