using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace AnimationClassLibrary
{
    /// <summary>
    /// ؼֵ֮֡ģʽ
    /// </summary>
    public enum LerpMode
    {
        /// <summary>
        /// Բֵ
        /// </summary>
        Linear,
        /// <summary>
        /// ֵ߲
        /// </summary>
        Curve,
    }
    
    /// <summary>
    /// ڲAnimationClip͵Ķݣ
    /// 任󣬻Եŵٶȡ
    /// </summary>
    public class AnimationPlayer
    {
        Model model;                           // ҪʩӶģ
        Matrix[] originalModelTransforms;      // ģ͵ԭʼ任
        
        TimeSpan startTime, endTime;           // ſʼʱ̺ͽʱ
        bool loop;                             // Ƿѭ
        bool paused;                           // Ƿͣ״̬       
        protected float playbackRate = 1.0f;   // űʣ1ʾٶȲ
        LerpMode lerpMode;                     // ݵĲֵģʽ

        // ǰʱ̵λáת
        Vector3 position, scale;
        Quaternion rotation;
        protected Matrix transform;            // 任

        /// <summary>
        /// ǰŵĶƬ
        /// </summary>
        public AnimationClip CurrentClip { get; private set; } 

        /// <summary>
        /// ǷѾ
        /// </summary>
        public bool Done { get; private set; }

        /// <summary>
        /// ȡǰѾŵʱ
        /// </summary>
        public TimeSpan ElapsedTime{ get;private  set; }                  

        /// <summary>
        /// ȡǰŵĹؼ֡
        /// </summary>
        public int CurrentKeyframeIndex{ get; private set; }

        /// <summary>
        /// һAnimationClipͶݵĲ
        /// </summary>
        public AnimationPlayer(ref Model model)
        {
            this.model = model;
            originalModelTransforms = new Matrix[model.Bones.Count];
            model.CopyBoneTransformsTo(originalModelTransforms);
        }

        /// <summary>
        /// ¶
        /// </summary>
        public void Update(GameTime gameTime)
        {
            if (CurrentClip == null || Done)
                return;            
            if (paused)
                return;

            TimeSpan time = gameTime.ElapsedGameTime;
            // ٶ
            if (playbackRate != 1.0f)
                time = TimeSpan.FromSeconds(time.TotalSeconds * playbackRate);
            ElapsedTime += time;

            // вֵ
            // ŵʱѾָʱ...
            while (ElapsedTime >= (endTime - startTime))
            {
                // ѭ򽫶ʱ˵0
                if (loop)
                {
                    ElapsedTime -= (endTime - startTime);
                    CurrentKeyframeIndex = 0;
                }
                // 򣬽ʱнȡ
                else
                {
                    Done = true;
                    ElapsedTime = endTime - startTime;
                    break;
                }
            }

            // ȡؼ֡
            IList<Keyframe> keyframes = CurrentClip.Keyframes;

            TimeSpan totalTime = ElapsedTime + startTime;
            // ȡǰؼ֡
            while (keyframes[CurrentKeyframeIndex + 1].Time < totalTime)
                CurrentKeyframeIndex++;
            // ȡӵǰ֡ʼĶʱ
            totalTime -= keyframes[CurrentKeyframeIndex].Time;

            // ȡŵǰؼ֡һؼ֮֡λã01֮
            float amt = (float)((totalTime.TotalSeconds) /
                (keyframes[CurrentKeyframeIndex + 1].Time - 
                keyframes[CurrentKeyframeIndex].Time).TotalSeconds);
            // ݲֵģʽλýвֵ
            if (lerpMode == LerpMode.Linear)
                position = Vector3.Lerp(keyframes[CurrentKeyframeIndex].Position,
                    keyframes[CurrentKeyframeIndex + 1].Position, amt);
            else
                position = catmullRom3D(
                    keyframes[wrap(CurrentKeyframeIndex - 1, keyframes.Count - 1)].Position,
                    keyframes[wrap(CurrentKeyframeIndex, keyframes.Count - 1)].Position,
                    keyframes[wrap(CurrentKeyframeIndex + 1, keyframes.Count - 1)].Position,
                    keyframes[wrap(CurrentKeyframeIndex + 2, keyframes.Count - 1)].Position, amt);

            // ˶еתȻԲֵ
            rotation = Quaternion.Slerp(keyframes[CurrentKeyframeIndex].Rotation,
                keyframes[CurrentKeyframeIndex + 1].Rotation, amt);
            scale = Vector3.Lerp(keyframes[CurrentKeyframeIndex].Scale,
                keyframes[CurrentKeyframeIndex + 1].Scale, amt);

            // 任
            // 任
            transform = Matrix.CreateScale(scale) *
                Matrix.CreateFromQuaternion(rotation) *
                Matrix.CreateTranslation(position);

            // ȡǰŵĹؼ֡BoneȻֵıӦBoneı任
            model.Bones[CurrentClip.Keyframes[CurrentKeyframeIndex].BoneIndex].Transform = transform;
        }

        Vector3 catmullRom3D(Vector3 v1, Vector3 v2, Vector3 v3, Vector3 v4, float amt)
        {
            return new Vector3(
                MathHelper.CatmullRom(v1.X, v2.X, v3.X, v4.X, amt),
                MathHelper.CatmullRom(v1.Y, v2.Y, v3.Y, v4.Y, amt),
                MathHelper.CatmullRom(v1.Z, v2.Z, v3.Z, v4.Z, amt));
        }

        // "value"[0, max]Χѭ
        int wrap(int value, int max)
        {
            while (value > max)
                value -= max;

            while (value < 0)
                value += max;

            return value;
        }

        /// <summary>
        /// ĶƬ
        /// </summary>
        /// <param name="clip">ҪŵĶƬ</param>
        /// <param name="loop">Ƿѭ</param>
        /// <param name="lerpMode">任ֵģʽ</param>
        /// <param name="playbackRate">ű</param>
        public void StartClip(AnimationClip clip, bool loop, LerpMode lerpMode, float playbackRate)
        { 
            StartClip(clip, TimeSpan.FromSeconds(0), clip.Duration,
                loop, lerpMode, playbackRate);
        }

        /// <summary>
        /// ݸĹؼ֡ؼ֮֡ĶƬ
        /// </summary>
        /// <param name="clip">ҪŵĶƬ</param>
        /// <param name="startFrame">ʼʱ̵Ĺؼ֡</param>
        /// <param name="endFrame">ʱ̵Ĺؼ֡</param>
        /// <param name="loop">Ƿѭ</param>
        /// <param name="lerpMode">任ֵģʽ</param>
        /// <param name="playbackRate">ű</param>
        public void StartClip(AnimationClip clip, int startFrame, int endFrame, 
            bool loop, LerpMode lerpMode, float playbackRate)
        {
            StartClip(clip, clip.Keyframes[startFrame].Time,
                clip.Keyframes[endFrame].Time,
                loop, lerpMode, playbackRate);
        }

        /// <summary>
        /// ݸĿʼʱ̺ͽʱ̲һʱڵĶƬ
        /// </summary>
        /// <param name="clip">ҪŵĶƬ</param>
        /// <param name="StartTime">ʼʱ</param>
        /// <param name="EndTime">ʱ</param>
        /// <param name="loop">Ƿѭ</param>
        /// <param name="lerpMode">任ֵģʽ</param>
        /// <param name="playbackRate">ű</param>
        public void StartClip(AnimationClip  clip, TimeSpan StartTime, TimeSpan EndTime, 
            bool loop,LerpMode lerpMode, float playbackRate)
        {
            this.CurrentClip = clip;
            this.startTime = StartTime;
            this.endTime = EndTime;
            this.loop = loop;
            this.lerpMode = lerpMode;
            this.playbackRate = playbackRate;

            ElapsedTime = TimeSpan.FromSeconds(0);
            CurrentKeyframeIndex = 0;

            transform = Matrix.Identity;
            Done = false;
        }

        /// <summary>
        /// ͣ
        /// </summary>
        public void PauseClip()
        {
            paused = true;
        }

        /// <summary>
        /// 
        /// </summary>
        public void ResumeClip()
        {
            paused = false;
        }
    }
}
