﻿using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace AnimationClassLibrary
{
    public class ModelAnimationBlender
    {
        struct animationProperty
        {
            public string animationName;
            public bool loop;
            public float playBackRate;
            public int startIndex;
            public int endIndex;
        }
        
        ModelAnimationData animationData;          // 动画数据
        RigidAnimationPlayer[] rigidPlayers;       // 播放器数组,包含两个播放器类
        int animationClipCount;                    // 当前混合的动画片段数量
        animationProperty animationProperty0;
        animationProperty animationProperty1;
        
        public ModelAnimationBlender(ModelAnimationData animationData,ref Model model)
        {
            // 初始化动画数据和播放器类
            this.animationData = animationData;
            animationClipCount = 0;
            animationProperty0 = new animationProperty();
            animationProperty1 = new animationProperty();
            
            rigidPlayers = new RigidAnimationPlayer[2];
            rigidPlayers[0] = new RigidAnimationPlayer(ref model);
            rigidPlayers[1] = new RigidAnimationPlayer(ref model);
        }

        public void Blend(string animationClipName, bool loop, float playBackRate)
        {
            this.Blend(animationClipName, loop, playBackRate, 0, animationData.ModelAnimationClips[animationClipName].Keyframes.Count - 1);
        }

        public void Blend(string newAnimationClipName, bool loop, float playBackRate, int startIndex, int endIndex)
        {
            if (!animationData.ModelAnimationClips.ContainsKey(newAnimationClipName))
                return;

            // 如果目前还没有播放动画，则使用第一个播放器播放新加入的动画片段
            if (this.animationClipCount == 0)
            {
                animationProperty0.animationName = newAnimationClipName;
                animationProperty0.loop = loop;
                animationProperty0.playBackRate = playBackRate;
                animationProperty0.startIndex = startIndex;
                animationProperty0.endIndex = endIndex;
                
                rigidPlayers[0].StartClip(animationData.ModelAnimationClips[newAnimationClipName],
                        loop, playBackRate, startIndex, endIndex);
                animationClipCount++;
            }
            // 如果目前播放了一个动画，则使用第二个播放器播放新加入的动画片段
            else if (this.animationClipCount == 1)
            {
                animationProperty1.animationName = newAnimationClipName;
                animationProperty1.loop = loop;
                animationProperty1.playBackRate = playBackRate;
                animationProperty1.startIndex = startIndex;
                animationProperty1.endIndex = endIndex;
                
                rigidPlayers[1].StartClip(animationData.ModelAnimationClips[newAnimationClipName],
                        loop, playBackRate, startIndex, endIndex);
                animationClipCount++;
            }
            // 如果现在已经播放了两个动画片段
            else if (this.animationClipCount == 2)
            {
                // 则在动画片段数组中移除第一个动画，将第二个动画作为第一个动画，
                // 将新加入的动画作为第二个动画
                animationProperty0.animationName = animationProperty1.animationName;
                animationProperty0.loop = animationProperty1.loop;
                animationProperty0.playBackRate = animationProperty1.playBackRate;
                animationProperty0.startIndex = animationProperty1.startIndex;
                animationProperty0.endIndex = animationProperty1.endIndex;

                animationProperty1.animationName = newAnimationClipName;
                animationProperty1.loop = loop;
                animationProperty1.playBackRate = playBackRate;
                animationProperty1.startIndex = startIndex;
                animationProperty1.endIndex = endIndex;

                rigidPlayers[0].StartClip(animationData.ModelAnimationClips[animationProperty0.animationName],
                        animationProperty0.loop, animationProperty0.playBackRate, animationProperty0.startIndex, animationProperty0.endIndex);
                rigidPlayers[1].StartClip(animationData.ModelAnimationClips[newAnimationClipName],
                        loop, playBackRate, startIndex, endIndex);
            }
        }

        /// <summary>
        /// 更新动画数据
        /// </summary>
        public void Update(GameTime gameTime)
        {
            rigidPlayers[0].Update(gameTime);            
            rigidPlayers[1].Update(gameTime);
        }
    }
}
