using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;

namespace AnimationClassLibrary
{
    /// <summary>
    /// Ƥģ͵Ķ࣬ŵSkinningAnimationData͵Ķݡ
    /// </summary>
    public class SkinnedAnimationPlayer : ModelAnimationPlayerBase
    {        
        // ǰĶ任
        Matrix[] boneTransforms;
        Matrix[] worldTransforms;
        Matrix[] skinTransforms;

        List<Matrix> bindPose;
        List<Matrix> inverseBindPose;
        List<int> skeletonHierarchy;
        
        /// <summary>
        /// һƤģͶࡣ
        /// </summary>
        public SkinnedAnimationPlayer(List<Matrix> bindPose, List<Matrix> inverseBindPose, List<int> skeletonHierarchy)
        {
            if (bindPose == null || bindPose.Count == 0)
                throw new Exception("");
            
            boneTransforms = new Matrix[bindPose.Count];
            worldTransforms = new Matrix[bindPose.Count];
            skinTransforms = new Matrix[bindPose.Count];

            this.bindPose = bindPose;
            this.inverseBindPose = inverseBindPose;
            this.skeletonHierarchy = skeletonHierarchy;                
        }

        /// <summary>
        /// ʼƬ
        /// </summary>
        protected override void InitClip()
        {
            bindPose.CopyTo(boneTransforms);
        }

        /// <summary>
        /// ؼ֡õboneTransformӦλá
        /// </summary>
        protected override void SetKeyframe(ModelKeyframe keyframe)
        {
            boneTransforms[keyframe.BoneIndex] = keyframe.Transform;
        }

        /// <summary>
        /// Updates the transformations ultimately needed for rendering
        /// </summary>
        protected override void OnUpdate()
        {
            if (CurrentClip != null)
            {
                // ڵ任
                worldTransforms[0] = boneTransforms[0];
                skinTransforms[0] = inverseBindPose[0] * worldTransforms[0];

                // ӽڵ任
                for (int bone = 1; bone < worldTransforms.Length; bone++)
                {
                    int parentBone = skeletonHierarchy[bone];

                    worldTransforms[bone] = boneTransforms[bone] * worldTransforms[parentBone];
                    skinTransforms[bone] = inverseBindPose[bone] * worldTransforms[bone];
                }                
            }
        }                       

        /// <summary>
        /// Gets the current bone transform matrices, relative to their parent bones.
        /// </summary>
        private Matrix[] GetBoneTransforms()
        {
            return boneTransforms;
        }

        /// <summary>
        /// Gets the current bone transform matrices, in absolute format.
        /// </summary>
        private Matrix[] GetWorldTransforms()
        {
            return worldTransforms;
        }

        /// <summary>
        /// Gets the current bone transform matrices, relative to the skinning bind pose.
        /// </summary>
        public Matrix[] GetSkinTransforms()
        {
            return skinTransforms;
        }
    }
}
