using System;
using System.Collections.Generic;
using AnimationClassLibrary;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline.Processors;

namespace AnimationPipeline
{
    /// <summary>
    /// ԶݴӿõModelProcessor̳,
    /// ˶֧֡
    /// </summary>
    [ContentProcessor(DisplayName = "Animated Model Processor")]
    public class AnimatedModelProcessor : ModelProcessor
    {        
        /// <summary> 
        /// ProcessݹܵмʽNodeContentתΪݵModelContent
        /// </summary>
        public override ModelContent Process(NodeContent input, ContentProcessorContext context) 
        { 
            // ģͽм飬ȷĶǿԱݴд 
            ValidateMesh(input, context, null);
             
            // ModelProcessorȡģ 
            ModelContent model = base.Process(input, context);
            
            // ģڲĶƬμ
            Dictionary<string, ModelAnimationClip> modelAnimationClips = new Dictionary<string, ModelAnimationClip>();

            // Զݽд
            ProcessAnimations(input, model, modelAnimationClips);
                        
            // ݴ洢ģ͵Tag
            model.Tag = new ModelAnimationData(modelAnimationClips);

            return model;
        }


        /// <summary>
        /// ݹܵУAnimationContentʽģ
        /// XNAУAnimationClipʽģ
        /// ǽ˸ʽתĹ
        /// </summary>
        static void ProcessAnimations(
            NodeContent input,
            ModelContent model,
            Dictionary<string, ModelAnimationClip> modelAnimationClips)
        {            
            // һ˱boneӳ䵽
            // NameΪnullbone
            Dictionary<string, int> boneMap = new Dictionary<string, int>();
            for (int i = 0; i < model.Bones.Count; i++)
            {
                string boneName = model.Bones[i].Name;
            
                if (!string.IsNullOrEmpty(boneName))
                    boneMap.Add(boneName, i);
            }

            // ȡģڵĶȻǷmodelAnimationClips           
            foreach (KeyValuePair<string, AnimationContent> animation in input.Animations)
            {
                ModelAnimationClip processed = ProcessRootAnimation(animation.Value, model.Bones[0].Name);
                modelAnimationClips.Add(animation.Key, processed);
            }

            // ڵ֮еĶƣanimationNames
            List<string> animationNames = new List<string>();
            AddAnimationNodes(animationNames, input);

            // ȡڵµжȻǷmodelAnimationClips
            foreach (string key in animationNames)
            {
                ModelAnimationClip processed = ProcessAnimation(key, boneMap, input, model);    
                
                // keyͬkeyͬModelAnimationClipкϲ
                if(modelAnimationClips .ContainsKey (key))
                    modelAnimationClips[key].Merge(processed);
                else
                    modelAnimationClips.Add(key, processed);
            }
        }

        /// <summary>
        /// ȡڵ֮еĶ
        /// </summary>
        static void AddAnimationNodes(List<string> animationNames, NodeContent node)
        {            
            foreach (NodeContent childNode in node.Children)
            {
                // ڵ㶯ݲؼ֡ڵ
                foreach (string key in childNode.Animations.Keys)
                {
                    if (!animationNames.Contains(key))
                        animationNames.Add(key);
                }
                AddAnimationNodes(animationNames, childNode);
            }
        }

        /// <summary>
        /// ݹܵУAnimationContentʽģ
        /// XNAУAnimationClipʽģ
        /// ǽ˸ʽתĹ
        /// </summary>
        public static ModelAnimationClip ProcessRootAnimation(AnimationContent animation, string name)
        {
            List<ModelKeyframe> keyframes = new List<ModelKeyframe>();

            // animationΪAnimationContentһΪChannelsΪAnimationChannelDictionary
            // ԣAnimationChannelDictionaryһDictionaryļǶƣֵΪAnimationChannel
            // Ĺؼ֡ݣAnimationChannel̳ICollection<AnimationKeyframe>
            
            // ڵ㶯boneĸڵ
            AnimationChannel channel = animation.Channels[name];
            
            // 䵱ǰڵ㶯Ĺؼ֡
            foreach (AnimationKeyframe keyframe in channel)
            {
                keyframes.Add(new ModelKeyframe(0, keyframe.Time, keyframe.Transform));
            }            

            // ݹؼ֡ʱ
            keyframes.Sort(CompareKeyframeTimes);

            if (keyframes.Count == 0)
                throw new InvalidContentException("Animationûаؼ֡");

            if (animation.Duration <= TimeSpan.Zero)
                throw new InvalidContentException("AnimationʱΪ㡣");

            return new ModelAnimationClip(animation.Duration, keyframes);
        }


        /// <summary>
        /// ݹܵУAnimationContentʽģ
        /// XNAУAnimationClipʽģ
        /// ǽ˸ʽתĹ
        /// </summary>
        static ModelAnimationClip ProcessAnimation(
            string animationName,
            Dictionary<string, int> boneMap,
            NodeContent input,
            ModelContent model)
        {
            List<ModelKeyframe> keyframes = new List<ModelKeyframe>();
            TimeSpan duration = TimeSpan.Zero;

            // ȡӽڵж
            AddTransformationNodes(animationName, boneMap, input, keyframes, ref duration);

            // ݹؼ֡ʱ
            keyframes.Sort(CompareKeyframeTimes);

            if (keyframes.Count == 0)
                throw new InvalidContentException("Animationûаؼ֡");

            if (duration <= TimeSpan.Zero)
                throw new InvalidContentException("AnimationʱΪ㡣");

            return new ModelAnimationClip(duration, keyframes);
        }

        /// <summary>
        /// ȡӽڵж
        /// </summary>
        static void AddTransformationNodes(
            string animationName, 
            Dictionary<string, int> boneMap, 
            NodeContent input, 
            List<ModelKeyframe> keyframes, 
            ref TimeSpan duration)
        {
            // ӽڵ
            foreach (NodeContent childNode in input.Children)
            {
                // ǰӽڵ㲻򲻽к̴
                if (childNode.Animations.ContainsKey(animationName))
                {
                    AnimationChannel childChannel = childNode.Animations[animationName].
                        Channels[childNode.Name];
                    // ȡʱֵΪʱ
                    if (childNode.Animations[animationName].Duration != duration)
                    {
                        if (duration < childNode.Animations[animationName].Duration)
                            duration = childNode.Animations[animationName].Duration;
                    }

                    int boneIndex;
                    if (!boneMap.TryGetValue(childNode.Name, out boneIndex))
                    {
                        throw new InvalidContentException(string.Format(
                            "bone '{0}'ҵģ͵һ.", 
                            childNode.Name));
                    }

                    foreach (AnimationKeyframe keyframe in childChannel)
                    {
                        keyframes.Add(new ModelKeyframe(boneIndex, keyframe.Time, keyframe.Transform));
                    }
                }

                AddTransformationNodes(animationName, boneMap, childNode, keyframes, ref duration);
            }
        }

        /// <summary>
        /// ڸʱеComparison
        /// </summary>
        static int CompareKeyframeTimes(ModelKeyframe a, ModelKeyframe b)
        {
            return a.Time.CompareTo(b.Time);
        }


        /// <summary>
        /// ģͽм飬ȷĶǿԴġ 
        /// ˵MeshContentĸڵΪһBoneContentNameΪnull,
        /// ǲҪģͣ޷
        /// </summary>
        static void ValidateMesh(NodeContent node, ContentProcessorContext context, string parentBoneName)
        {
            MeshContent mesh = node as MeshContent;
            
            // ֤mesh
            if (mesh != null)
            {
                // ֤mesh
                if (parentBoneName != null)
                {
                    context.Logger.LogWarning(null, null,
                        "Mesh {0} bone {1}ӽڵ㣬AnimatedModelProcessor " +
                        "޷ȷmesh",
                        mesh.Name, parentBoneName);
                }               
            }
            else if (node is BoneContent)
            {
                // һBoneڵ, 򱣴ںĵݹ̡
                parentBoneName = node.Name;
            }

            // ݹڵнڵ
            foreach (NodeContent child in new List<NodeContent>(node.Children))
                ValidateMesh(child, context, parentBoneName);
        }
    }
}
