using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using AnimationClassLibrary;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content.Pipeline;

namespace AnimationPipeline
{
    /// <summary>
    /// ཫAnimationDataתΪModelAnimationDataݲԶвֵ㡣
    /// </summary>
    [ContentProcessor(DisplayName = "AnimationData Processor")]
    public class AnimationDataProcessor : ContentProcessor<AnimationData, ModelAnimationData>
    {
        private float rate = 30.0f;

        [DisplayName("Rate")]
        [DefaultValue(30.0f)]
        [Description("ֵĲƵʣĬΪÿ30βֵ㡣")]
        public float Rate
        {
            get { return rate; }
            set { rate = value; }
        }

        LerpMode lerpMode = LerpMode.Linear;

        [DisplayName("LerpMode")]
        [DefaultValue(LerpMode.Linear)]
        [Description("ֵģʽΪԲֵֵ߲ĬΪԲֵ")]
        public LerpMode LerpMode
        {
            get { return lerpMode; }
            set { lerpMode = value; }
        }       
        
        public override ModelAnimationData Process(AnimationData input, ContentProcessorContext context)
        {
            Dictionary<string, ModelAnimationClip> modelAnimationClips = new Dictionary<string,ModelAnimationClip>();
 
            // ModelAnimationClipsеĶ       
            foreach (KeyValuePair<string, AnimationClip> animationDictionary in input.AnimationClips)
            {
                ModelAnimationClip modelAnimationClip;
                ConvertAnimationData(animationDictionary.Value , out modelAnimationClip);

                modelAnimationClips.Add(animationDictionary.Key , modelAnimationClip); 
            }

            return new ModelAnimationData(modelAnimationClips);
        }

        /// <summary>
        /// AnimationClip͵ĶתΪModelAnimationClip͵Ķ
        /// </summary>
        private void ConvertAnimationData(AnimationClip animationClip, out ModelAnimationClip modelAnimationClip)
        {
            // ȡǰAnimationClipеĹؼ֡
            List<Keyframe> keyframes = animationClip.Keyframes;
            // ؼ֡ҪΰBoneIndexTimeԽ
            // ˴ʹLinq﷨Ϊࡣ
            var query = from keyframe in keyframes
                        orderby keyframe.BoneIndex ,keyframe.Time 
                    select keyframe;
            keyframes = query.ToList();
            // Ĺؼ֡ϣΪModelKeyframeĹҪµ
            // С
            List<ModelKeyframe> outKeyframes = new List<ModelKeyframe>();
            
            
            int currentKeyframeIndex = 0;
            // ֶӵһ֡Ĺؼ֡
            AddKeyframe(keyframes[0], outKeyframes);

            // йؼ֡
            while (currentKeyframeIndex < keyframes.Count - 1)
            {
                Keyframe currentKeyframe = keyframes[currentKeyframeIndex];
                Keyframe nextKeyframe = keyframes[currentKeyframeIndex + 1];

                // ǰؼ֡BoneIndexһ֮֡ͬܶвֵ
                // 򶯻ǲȷġ
                if (currentKeyframe.BoneIndex != nextKeyframe.BoneIndex)
                {
                    // ֱһ֡Ĺؼ֡Ȼһѭ
                    AddKeyframe(nextKeyframe, outKeyframes);
                    currentKeyframeIndex++;
                    continue;
                }

                // һЩʱ
                Vector3 position, scale;
                Quaternion rotation;
                TimeSpan ElapsedTime = TimeSpan.Zero;
                TimeSpan Interval = nextKeyframe.Time - currentKeyframe.Time;

                #region ڵǰؼ֡һؼ֮֡вֵ

                while (ElapsedTime < Interval)
                {
                    ElapsedTime += TimeSpan.FromSeconds(1 / rate);

                    // Ѿһؼ֡,ֶһؼ֡
                    if (ElapsedTime > Interval)
                    {
                        AddKeyframe(nextKeyframe, outKeyframes);
                        // while (ElapsedTime < Interval)
                        break;
                    }

                    // ȡŵǰؼ֡һؼ֮֡λã01֮
                    float amt = (float)(ElapsedTime.TotalSeconds / Interval.TotalSeconds);
                    // ݲֵģʽλýвֵ
                    if (lerpMode == LerpMode.Linear)
                        position = Vector3.Lerp(currentKeyframe.Position,
                            nextKeyframe.Position, amt);
                    else
                        position = catmullRom3D(
                            keyframes[wrap(currentKeyframeIndex - 1, keyframes.Count - 1)].Position,
                            keyframes[wrap(currentKeyframeIndex, keyframes.Count - 1)].Position,
                            keyframes[wrap(currentKeyframeIndex + 1, keyframes.Count - 1)].Position,
                            keyframes[wrap(currentKeyframeIndex + 2, keyframes.Count - 1)].Position, amt);

                    // ˶еתȻԲֵ
                    rotation = Quaternion.Slerp(currentKeyframe.Rotation,
                        nextKeyframe.Rotation, amt);
                    scale = Vector3.Lerp(currentKeyframe.Scale,
                        nextKeyframe.Scale, amt);

                    // 任
                    Matrix transform = Matrix.CreateScale(scale) *
                        Matrix.CreateFromQuaternion(rotation) *
                        Matrix.CreateTranslation(position);

                    ModelKeyframe keyframe = new ModelKeyframe(currentKeyframe .BoneIndex , ElapsedTime + currentKeyframe.Time, transform);
                    outKeyframes.Add(keyframe);
                }
                #endregion

                currentKeyframeIndex++;
            }

            // ϻҪModelKeyframeϰTime
            outKeyframes.Sort(CompareKeyframeTimes);
            modelAnimationClip = new ModelAnimationClip(animationClip.Duration, outKeyframes);
        }

        // ؼ֡뵽ModelKeyframe
        private void AddKeyframe(Keyframe keyframe, List<ModelKeyframe> outKeyframes)
        {
            // 任
            Matrix transform = Matrix.CreateScale(keyframe.Scale) *
                Matrix.CreateFromQuaternion(keyframe.Rotation) *
                Matrix.CreateTranslation(keyframe.Position);
            ModelKeyframe ModelKeyframe = new ModelKeyframe(keyframe.BoneIndex, keyframe.Time, transform);
            outKeyframes.Add(ModelKeyframe);
        }

        // չؼ֡TimeComparison
        private static int CompareKeyframeTimes(ModelKeyframe a, ModelKeyframe b)
        {
            return a.Time.CompareTo(b.Time);
        }

        // ֵ߲
        Vector3 catmullRom3D(Vector3 v1, Vector3 v2, Vector3 v3, Vector3 v4, float amt)
        {
            return new Vector3(
                MathHelper.CatmullRom(v1.X, v2.X, v3.X, v4.X, amt),
                MathHelper.CatmullRom(v1.Y, v2.Y, v3.Y, v4.Y, amt),
                MathHelper.CatmullRom(v1.Z, v2.Z, v3.Z, v4.Z, amt));
        }

        // "value"[0, max]Χѭ
        int wrap(int value, int max)
        {
            while (value > max)
                value -= max;

            while (value < 0)
                value += max;

            return value;
        }
    }
}