using System;
using System.Collections.Generic;
using System.ComponentModel;
using AnimationClassLibrary;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline.Processors;

namespace AnimationPipeline
{
    /// <summary>
    /// XNAеModelProcessorչƤ֧ 
    /// </summary>
    [ContentProcessor(DisplayName = "Skinned Model Processor")]
    public class SkinnedModelProcessor : ModelProcessor
    {
        // shader 2.0һpassԻƵbone
        // ıֵҪSkinnedModel.fxжӦֵ
        const int MaxBones = 59;

        /// <summary>
        /// дProcess˷ݹܵNodeContentмʽתΪݵContent
        /// </summary>
        public override ModelContent Process(NodeContent input,ContentProcessorContext context)
        {
            // ȷmesh֪δ
            ValidateMesh(input, context, null);

            // ȡĸbone
            BoneContent skeleton = MeshHelper.FindSkeleton(input);

            if (skeleton == null)
                throw new InvalidContentException("δҵڵskeleton");

            // We don't want to have to worry about different parts of the model being
            // in different local coordinate systems, so let's just bake everything.
            // ԭʹBakeһʣΪ決Ϊ3DеһרʣĲǼÿboneľԱ任
            FlattenTransforms(input, skeleton);

            // skeletonĲνṹתΪһBoneContent
            IList<BoneContent> bones = MeshHelper.FlattenSkeleton(skeleton);

            if (bones.Count > MaxBones)
            {
                throw new InvalidContentException(string.Format("{0}boneֻ֧{1}",bones.Count, MaxBones));
            }

            List<Matrix> bindPose = new List<Matrix>();
            List<Matrix> inverseBindPose = new List<Matrix>();
            List<int> skeletonHierarchy = new List<int>();
            
            // ȡbind pose͹
            foreach (BoneContent bone in bones)
            {
                bindPose.Add(bone.Transform);
                inverseBindPose.Add(Matrix.Invert(bone.AbsoluteTransform));
                skeletonHierarchy.Add(bones.IndexOf(bone.Parent as BoneContent));
            }

            // תΪʵʱеAnimationClipʽ
            Dictionary<string, ModelAnimationClip> animationClips;
            animationClips = ProcessAnimations(skeleton.Animations, bones);

            // ûProcessתģ
            ModelContent model = base.Process(input, context);

            // Զ嶯ݴ洢ģ͵Tag
            model.Tag = new SkinningAnimationData(animationClips, bindPose, inverseBindPose, skeletonHierarchy);

            return model;
        }


        /// <summary>
        /// ݹܵAnimationContentDictionaryתΪʵʱеAnimationClipʽ
        /// </summary>
        static Dictionary<string, ModelAnimationClip> ProcessAnimations(AnimationContentDictionary animations, IList<BoneContent> bones)
        {
            // һdictionaryboneƶӦbone
            Dictionary<string, int> boneMap = new Dictionary<string, int>();

            // dictionary
            for (int i = 0; i < bones.Count; i++)
            {
                string boneName = bones[i].Name;

                if (!string.IsNullOrEmpty(boneName))
                    boneMap.Add(boneName, i);
            }

            // תÿƬ
            Dictionary<string, ModelAnimationClip> animationClips;
            animationClips = new Dictionary<string, ModelAnimationClip>();

            foreach (KeyValuePair<string, AnimationContent> animation in animations)
            {
                ModelAnimationClip processed = ProcessAnimation(animation.Value, boneMap);

                animationClips.Add(animation.Key, processed);
            }

            if (animationClips.Count == 0)
            {
                throw new InvalidContentException("ļκζ");
            }

            return animationClips;
        }


        /// <summary>
        /// ݹܵAnimationContentDictionaryתΪʵʱеAnimationClipʽ
        /// </summary>
        static ModelAnimationClip ProcessAnimation(AnimationContent animation, Dictionary<string, int> boneMap)
        {
            List<ModelKeyframe> keyframes = new List<ModelKeyframe>();

            // ÿͨ
            foreach (KeyValuePair<string, AnimationChannel> channel in
                animation.Channels)
            {
                // ȡǰͨƵbone
                int boneIndex;

                if (!boneMap.TryGetValue(channel.Key, out boneIndex))
                {
                    throw new InvalidContentException(string.Format(
                        "δbone'{0}'ҵ, " +
                        "ǹһ֡", channel.Key));
                }

                // תؼ֡
                foreach (AnimationKeyframe keyframe in channel.Value)
                {
                    keyframes.Add(new ModelKeyframe(boneIndex, keyframe.Time, keyframe.Transform));
                }
            }

            // ʱԹؼ֡
            keyframes.Sort(CompareKeyframeTimes);

            if (keyframes.Count == 0)
                throw new InvalidContentException("ؼ֡");

            if (animation.Duration <= TimeSpan.Zero)
                throw new InvalidContentException("ʱΪ0");

            return new ModelAnimationClip(animation.Duration, keyframes);
        }


        /// <summary>
        /// ؼ֡ʱ
        /// </summary>
        static int CompareKeyframeTimes(ModelKeyframe a, ModelKeyframe b)
        {
            return a.Time.CompareTo(b.Time);
        }

        /// <summary>
        /// ȷmesh֪δ
        /// </summary>
        static void ValidateMesh(NodeContent node, ContentProcessorContext context, string parentBoneName)
        {
            MeshContent mesh = node as MeshContent;

            if (mesh != null)
            {
                // ȷmesh.
                if (parentBoneName != null)
                {
                    context.Logger.LogWarning(null, null, "Mesh {0}bone {1}һӽڵ㣬SkinnedModelProcessor " +
                        "޷ȷΪboneӽڵmesh", mesh.Name, parentBoneName);
                }

                if (!MeshHasSkinning(mesh))
                {
                    context.Logger.LogWarning(null, null,"Mesh {0} skinningϢԻᱻɾ",mesh.Name);
                    mesh.Parent.Children.Remove(mesh);
                    return;
                }
            }
            else if (node is BoneContent)
            {
                // ˽ڵһBoneContentҪڽһ顣
                parentBoneName = node.Name;
            }

            // ݹӽڵ (Ҫӽڵ㼯ϵĸΪпValidateMeshлɾĳЩӽڵ)
            foreach (NodeContent child in new List<NodeContent>(node.Children))
                ValidateMesh(child, context, parentBoneName);
        }


        /// <summary>
        /// meshǷskininngϢ
        /// </summary>
        static bool MeshHasSkinning(MeshContent mesh)
        {
            foreach (GeometryContent geometry in mesh.Geometry)
            {
                if (!geometry.Vertices.Channels.Contains(VertexChannelNames.Weights()))
                    return false;
            }

            return true;
        }


        /// <summary>
        /// 決ģͼı任жͬһϵС
        /// </summary>
        static void FlattenTransforms(NodeContent node, BoneContent skeleton)
        {
            foreach (NodeContent child in node.Children)
            {
                // Ҫڵ㣬Ϊ
                if (child == skeleton)
                    continue;

                // 決ıر任
                MeshHelper.TransformScene(child, child.Transform);

                // 決ɺǾͿԽϵΪλ
                child.Transform = Matrix.Identity;

                // ݹ鴦ӽڵ
                FlattenTransforms(child, skeleton);
            }
        }

        /// <summary>
        /// ǿвʶʹSkinnedEffect
        /// </summary>
        [DefaultValue(MaterialProcessorDefaultEffect.SkinnedEffect)]
        public override MaterialProcessorDefaultEffect DefaultEffect
        {
            get { return MaterialProcessorDefaultEffect.SkinnedEffect; }
            set { }
        }
    }
}
